import dataclasses
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Iterable, List, Optional, TextIO, Type, Union

from rdflib import URIRef
from urlpath import URL

from iolanta.context import merge
from iolanta.conversions import path_to_url, url_to_path
from iolanta.ensure_is_context import NotAContext, ensure_is_context
from iolanta.loaders.base import Loader, SourceType
from iolanta.loaders.local_file import LocalFile
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.base import Parser
from mkdocs_iolanta.types import MKDOCS


def merge_contexts(*contexts: LDContext) -> LDContext:
    return reduce(
        merge,
        filter(bool, contexts),
        {},
    )


@dataclass(frozen=True)
class LocalDirectory(Loader[URL]):
    """
    Retrieve Linked Data from a file on local disk.

    Requires URL with file:// scheme as input.
    """

    context_filenames: List[str] = field(
        default_factory=lambda: [
            'context.yaml',
            'context.yml',
            'context.json',
        ],
    )

    def find_context(self, source: SourceType) -> LDContext:
        raise ValueError('?!!!???')

    def directory_level_context(self, path: Path) -> Optional[LDContext]:
        for file_name in self.context_filenames:
            if (context_path := path / file_name).is_file():
                document = LocalFile().as_jsonld_document(
                    source=path_to_url(context_path),
                )

                if document:
                    try:
                        return ensure_is_context(document)
                    except NotAContext as err:
                        raise dataclasses.replace(
                            err,
                            path=context_path,
                        )
        return None

    def choose_parser_class(self, source: URL) -> Type[Parser]:
        """Choose parser class based on file extension."""
        raise ValueError('This is a directory')

    def as_quad_stream(
        self,
        source: Union[URL, Path],
        iri: Optional[URIRef],
        root_loader: Loader[URL],
        context: Optional[LDContext] = None,
    ) -> Iterable[Quad]:
        """Extract a sequence of quads from a local file."""
        path = url_to_path(source) if isinstance(source, URL) else source

        if not path.is_dir():
            yield from LocalFile().as_quad_stream(
                source=source,
                root_loader=root_loader,
                iri=iri,
                context=context,
            )
            return

        context = merge_contexts(
            context,
            self.directory_level_context(path),
        )

        for child in path.iterdir():
            child_iri = URIRef(f'{iri}{child.name}')

            if child.is_dir():
                child_iri += '/'

                yield from LocalDirectory().as_quad_stream(
                    source=child,
                    iri=child_iri,
                    root_loader=root_loader,
                    context=context,
                )

            elif child.stem != 'context':
                yield from LocalFile().as_quad_stream(
                    source=path_to_url(child),
                    iri=child_iri,
                    root_loader=root_loader,
                    context=context,
                )

            if iri is not None:
                yield Quad(
                    subject=child_iri,
                    predicate=MKDOCS.isChildOf,
                    object=iri,
                    graph=URIRef(
                        'https://iolanta.tech/loaders/local-directory',
                    ),
                )

    def as_file(self, source: URL) -> TextIO:
        """Construct a file-like object."""
        path = url_to_path(source)
        with path.open() as text_io:
            return text_io

    def as_jsonld_document(
        self,
        source: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """As JSON-LD document."""
        raise ValueError('This is a directory.')
